// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)
//

#include "muduo/net/Buffer.h"

#include <errno.h>
#include <sys/uio.h>

#include "muduo/net/SocketsOps.h"

using namespace muduo;
using namespace muduo::net;

const char Buffer::kCRLF[] = "\r\n";

const size_t Buffer::kCheapPrepend;
const size_t Buffer::kInitialSize;

/*
    Level Trigger（水平触发）和Edge Trigger（边缘触发）是与事件驱动编程模型相关的两种触发方式，通常用于描述事件在非阻塞 I/O 模型中的处理方式。
    这两种触发方式主要用于描述事件发生时系统如何通知应用程序以及应用程序如何响应事件。

    Level Trigger（水平触发）：
    当事件发生并且仍然处于活动状态时，系统会不断地通知应用程序，直到事件处理完成或被标记为非活动状态。
    水平触发适用于需要持续关注某个事件状态变化的情况。应用程序需要不断检查事件状态，直到事件不再活动。
    对于 I/O 事件，水平触发意味着当文件描述符上有数据可读或可写时，系统会不断通知应用程序直到所有数据被读取或写入。

    Edge Trigger（边缘触发）：
    当事件状态发生变化时（例如从无数据到有数据，或从有数据到无数据），系统会通知应用程序一次。即只在状态发生变化时触发一次，而不会持续通知。
    边缘触发适用于只关心事件状态变化瞬间的情况。应用程序需要在每次事件触发时立即处理，而不是持续关注事件状态。
    对于 I/O 事件，边缘触发意味着只有在文件描述符状态变化时（例如从不可读到可读，或从不可写到可写），系统会通知应用程序一次。
    在较低层次的操作系统和网络编程中，通常通过系统调用如 select、poll、epoll（在Linux中）等来实现水平触发。
    而边缘触发通常需要更高级别的接口，例如 epoll 中的 EPOLLET 标志。

    选择使用水平触发还是边缘触发通常取决于应用程序的需求。水平触发允许应用程序以更简单的方式处理事件，但可能导致处理效率低下。
    边缘触发则更适用于一些对事件状态变化非常敏感的场景，但相对来说可能更复杂一些。

    Buffer::readFd() 只调用一次 read(2)，而没有反复调用 read(2) 直到其
    返回 EAGAIN。首先，这么做是正确的，因为 muduo 采用 level trigger，这么做不会
    丢失数据或消息。其次，对追求低延迟的程序来说，这么做是高效的，因为每次读数
    据只需要一次系统调用。再次，这样做照顾了多个连接的公平性，不会因为某个连接
    上数据量过大而影响其他连接处理消息。
*/

ssize_t Buffer::readFd(int fd, int* savedErrno)
{
    // saved an ioctl()/FIONREAD call to tell how much to read
    char         extrabuf[65536];
    struct iovec vec[2];
    const size_t writable = writableBytes();
    vec[0].iov_base = begin() + writerIndex_;
    vec[0].iov_len = writable;
    vec[1].iov_base = extrabuf;
    vec[1].iov_len = sizeof extrabuf;
    // when there is enough space in this buffer, don't read into extrabuf.
    // when extrabuf is used, we read 128k-1 bytes at most.
    const int     iovcnt = (writable < sizeof extrabuf) ? 2 : 1;
    const ssize_t n = sockets::readv(fd, vec, iovcnt);
    if (n < 0) {
        *savedErrno = errno;
    } else if (implicit_cast<size_t>(n) <= writable) {
        writerIndex_ += n;
    } else {
        writerIndex_ = buffer_.size();
        append(extrabuf, n - writable);
    }
    // if (n == writable + sizeof extrabuf)
    // {
    //   goto line_30;
    // }
    return n;
}
