// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)

#include "muduo/net/Channel.h"

#include <poll.h>

#include <sstream>

#include "muduo/base/Logging.h"
#include "muduo/net/EventLoop.h"

using namespace muduo;
using namespace muduo::net;

const int Channel::kNoneEvent = 0;
const int Channel::kReadEvent = POLLIN | POLLPRI;
const int Channel::kWriteEvent = POLLOUT;

Channel::Channel(EventLoop* loop, int fd__)
    : loop_(loop),
      fd_(fd__),
      events_(0),
      revents_(0),
      index_(-1),
      logHup_(true),
      tied_(false),
      eventHandling_(false),
      addedToLoop_(false)
{
}

Channel::~Channel()
{
    assert(!eventHandling_);
    assert(!addedToLoop_);
    if (loop_->isInLoopThread()) {
        assert(!loop_->hasChannel(this));
    }
}

void Channel::tie(const std::shared_ptr<void>& obj)
{
    tie_ = obj;
    tied_ = true;
}

void Channel::update()
{
    addedToLoop_ = true;
    loop_->updateChannel(this);
}

void Channel::remove()
{
    assert(isNoneEvent());
    addedToLoop_ = false;
    loop_->removeChannel(this);
}

/*
    Channel::handleEvent() 是 Channel 的核心，它由 EventLoop::loop() 调用，它
    的功能是根据 revents_ 的值分别调用不同的用户回调。
*/
void Channel::handleEvent(Timestamp receiveTime)
{
    std::shared_ptr<void> guard;
    if (tied_) {
        guard = tie_.lock();
        if (guard) {
            handleEventWithGuard(receiveTime);
        }
    } else {
        handleEventWithGuard(receiveTime);
    }
}

/*
    简单介绍一下：POLLOUT、POLLPRI 、POLLRDHUP 、POLLIN

    这些是用于 Linux 中的 poll 系统调用中的事件标志位，它们用于指定对文件描述符进行轮询时感兴趣的事件。以下是对它们的简单介绍：

    POLLOUT:
    意义：表示文件描述符是否可写。如果设置了该标志，表示你对文件描述符上的写操作感兴趣。
    示例用途：当你希望知道是否可以往某个套接字（socket）写入数据时，可以设置 POLLOUT 标志进行轮询。

    POLLPRI:
    意义：表示有紧急数据可读。这通常用于带外（out-of-band）数据，例如带外数据的套接字上的紧急指针被设置。
    示例用途：在带外数据到达时，你可能会设置 POLLPRI 标志进行轮询。

    POLLRDHUP:
    意义：表示对端关闭连接，或者半关闭的状态。如果设置了该标志，表示你对文件描述符的读操作感兴趣，并且对端关闭了连接（或者半关闭）。
    示例用途：在网络编程中，你可能希望在对端关闭连接时进行一些清理工作，可以设置 POLLRDHUP 标志进行轮询。

    POLLIN:
    意义：表示文件描述符是否可读。如果设置了该标志，表示你对文件描述符上的读操作感兴趣。
    示例用途：当你希望知道是否可以从某个套接字（socket）读取数据时，可以设置 POLLIN 标志进行轮询。
    这些标志位可通过位掩码进行组合，以满足对多个事件的关注。在 poll 系统调用的使用中，你会将关注的事件标志设置到 struct pollfd 结构体
    的 events 字段中，然后调用 poll 函数进行轮询。

    这段代码根据 revents_ 的值，调用了不同的用户注册的回调函数，以处理不同的事件。
    这是一个典型的事件驱动编程的实现，通过设置不同的回调函数，用户可以在发生特定事件时执行自定义的操作。
*/
void Channel::handleEventWithGuard(Timestamp receiveTime)
{
    eventHandling_ = true;  // 这里用于防止在处理事件的过程中被析构。
    LOG_TRACE << reventsToString();

    /*
        revents_ & POLLHUP 这种 与 的语法表示：
            如果发生了 POLLHUP 事件且没有发生 POLLIN 事件，说明连接被挂起（通常表示对端关闭连接）。
            在这种情况下，如果设置了 logHup_，则记录一个警告日志，然后调用 closeCallback_ 回调函数。
    */
    if ((revents_ & POLLHUP) && !(revents_ & POLLIN)) {
        if (logHup_) {
            LOG_WARN << "fd = " << fd_ << " Channel::handle_event() POLLHUP";
        }
        if (closeCallback_) closeCallback_();
    }

    /*
        revents_ & POLLNVAL: 如果发生了 POLLNVAL 事件，表示描述符没有打开，输出警告日志。
    */
    if (revents_ & POLLNVAL) {
        LOG_WARN << "fd = " << fd_ << " Channel::handle_event() POLLNVAL";
    }

    if (revents_ & (POLLERR | POLLNVAL)) {
        // 错误回调
        if (errorCallback_) errorCallback_();
    }
    if (revents_ & (POLLIN | POLLPRI | POLLRDHUP)) {
        // 读取回调
        if (readCallback_) readCallback_(receiveTime);
    }
    if (revents_ & POLLOUT) {
        if (writeCallback_) writeCallback_();
    }
    eventHandling_ = false;
}

string Channel::reventsToString() const
{
    return eventsToString(fd_, revents_);
}

string Channel::eventsToString() const { return eventsToString(fd_, events_); }

string Channel::eventsToString(int fd, int ev)
{
    std::ostringstream oss;
    oss << fd << ": ";
    if (ev & POLLIN) oss << "IN ";
    if (ev & POLLPRI) oss << "PRI ";
    if (ev & POLLOUT) oss << "OUT ";
    if (ev & POLLHUP) oss << "HUP ";
    if (ev & POLLRDHUP) oss << "RDHUP ";
    if (ev & POLLERR) oss << "ERR ";
    if (ev & POLLNVAL) oss << "NVAL ";

    return oss.str();
}
