// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)
//
// This is an internal header file, you should not include this.

#ifndef MUDUO_NET_CHANNEL_H
#define MUDUO_NET_CHANNEL_H

#include <functional>
#include <memory>

#include "muduo/base/Timestamp.h"
#include "muduo/base/noncopyable.h"

namespace muduo {
namespace net {

class EventLoop;

/*

    Channel 类用于表示一个文件描述符（fd）上的事件，这可以是套接字上的可读事件、可写事件等。

    每个 Channel 对象自始至终只属于一个 EventLoop，因此每个 Channel 对
    象都只属于某一个 IO 线程。每个 Channel
   对象自始至终只负责一个文件描述符（fd） 的 IO 事件分发，但它并不拥有这个
   fd，也不会在析构的时候关闭这个 fd。Channel 会把不同的 IO
   事件分发为不同的回调，例如 ReadCallback、WriteCallback 等，而且“回调”用
   boost::function 表示，用户无须继承 Channel，Channel 不是基类。 muduo
   用户一般不直接使用 Channel，而会使用更上层的封装，如 TcpConnection。 Channel
   的生命期由其 owner class 负责管理，它一般是其他 class 的直接或间接成 员。
*/

///
/// A selectable I/O channel.
///
/// This class doesn't own the file descriptor.
/// The file descriptor could be a socket,
/// an eventfd, a timerfd, or a signalfd
class Channel : noncopyable {
public:
    typedef std::function<void()>          EventCallback;
    typedef std::function<void(Timestamp)> ReadEventCallback;

    Channel(EventLoop* loop, int fd);
    ~Channel();

    /*
        Channel::handleEvent() 是 Channel 的核心，它由 EventLoop::loop() 调用，它
        的功能是根据 revents_ 的值分别调用不同的用户回调。
    */
    void handleEvent(Timestamp receiveTime);
    void setReadCallback(ReadEventCallback cb)
    {
        readCallback_ = std::move(cb);
    }
    void setWriteCallback(EventCallback cb) { writeCallback_ = std::move(cb); }
    void setCloseCallback(EventCallback cb) { closeCallback_ = std::move(cb); }
    void setErrorCallback(EventCallback cb) { errorCallback_ = std::move(cb); }

    /// Tie this channel to the owner object managed by shared_ptr,
    /// prevent the owner object being destroyed in handleEvent.
    void tie(const std::shared_ptr<void>&);

    int  fd() const { return fd_; }
    int  events() const { return events_; }
    void set_revents(int revt) { revents_ = revt; }  // used by pollers
    // int revents() const { return revents_; }
    bool isNoneEvent() const { return events_ == kNoneEvent; }

    void enableReading()
    {
        events_ |= kReadEvent;
        update();
    }
    void disableReading()
    {
        events_ &= ~kReadEvent;
        update();
    }
    void enableWriting()
    {
        events_ |= kWriteEvent;
        update();
    }
    void disableWriting()
    {
        events_ &= ~kWriteEvent;
        update();
    }
    void disableAll()
    {
        events_ = kNoneEvent;
        update();
    }
    bool isWriting() const { return events_ & kWriteEvent; }
    bool isReading() const { return events_ & kReadEvent; }

    // for Poller
    int  index() { return index_; }
    void set_index(int idx) { index_ = idx; }

    // for debug
    string reventsToString() const;
    string eventsToString() const;

    void doNotLogHup() { logHup_ = false; }

    EventLoop* ownerLoop() { return loop_; }
    void       remove();

private:
    static string eventsToString(int fd, int ev);

    /*
        Channel 的成员函数都只能在 IO 线程调用，因此更新数据成员都不必加锁。
    */
    void update();
    void handleEventWithGuard(Timestamp receiveTime);

    static const int kNoneEvent;
    static const int kReadEvent;
    static const int kWriteEvent;

    EventLoop* loop_;
    const int  fd_;
    int        events_;   // events_ 是它关心的 IO 事件, 由用户设置
    int        revents_;  // it's the received event types of epoll or poll

    /*
        revents_ 是目前活动的事件，由 EventLoop/Poller 设置；这两个字段都是 bit
        pattern，它们的名字来自 poll(2) 的 struct pollfd
    */

    int  index_;  // used by Poller.
    bool logHup_;

    std::weak_ptr<void> tie_;
    bool                tied_;
    bool                eventHandling_;
    bool                addedToLoop_;
    ReadEventCallback   readCallback_;
    EventCallback       writeCallback_;
    EventCallback       closeCallback_;
    EventCallback       errorCallback_;
};

}  // namespace net
}  // namespace muduo

#endif  // MUDUO_NET_CHANNEL_H
