// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)
//
// This is a public header file, it must only include public header files.

#ifndef MUDUO_NET_EVENTLOOP_H
#define MUDUO_NET_EVENTLOOP_H

#include <atomic>
#include <boost/any.hpp>
#include <functional>
#include <vector>

#include "muduo/base/CurrentThread.h"
#include "muduo/base/Mutex.h"
#include "muduo/base/Timestamp.h"
#include "muduo/net/Callbacks.h"
#include "muduo/net/TimerId.h"

namespace muduo {
namespace net {

class Channel;
class Poller;
class TimerQueue;

/*
    EventLoop 的构造函数会记住本对象所属的线程（threadId_）。
    创建了 EventLoop 对象的线程是 IO 线程，其主要功能是运行事件循环 EventLoop::
    loop()。
*/
///
/// Reactor, at most one per thread.
///
/// This is an interface class, so don't expose too much details.
class EventLoop : noncopyable {
public:
    typedef std::function<void()> Functor;

    EventLoop();
    ~EventLoop();  // force out-line dtor, for std::unique_ptr members.

    ///
    /// Loops forever.
    ///
    /// Must be called in the same thread as creation of the object.
    ///
    /*
        事件循环必须在 IO 线程执行，因此 EventLoop::loop() 会检查这一
       pre-condition。
    */
    void loop();

    /// Quits loop.
    ///
    /// This is not 100% thread safe, if you call through a raw pointer,
    /// better to call through shared_ptr<EventLoop> for 100% safety.
    void quit();

    ///
    /// Time when poll returns, usually means data arrival.
    ///
    Timestamp pollReturnTime() const { return pollReturnTime_; }

    int64_t iteration() const { return iteration_; }

    /*
        EventLoop 有一个非常有用的功能：在它的 IO 线程内执行某个用户任务回调，
        即 EventLoop::runInLoop(const Functor& cb)，其中 Functor 是
       boost::function<void()>。 如果用户在当前 IO
       线程调用这个函数，回调会同步进行；如果用户在其他线程调用 runInLoop()，cb
       会被加入队列，IO 线程会被唤醒来调用这个 Functor。
    */
    /// Runs callback immediately in the loop thread.
    /// It wakes up the loop, and run the cb.
    /// If in the same loop thread, cb is run within the function.
    /// Safe to call from other threads.
    void runInLoop(Functor cb);
    /// Queues callback in the loop thread.
    /// Runs after finish pooling.
    /// Safe to call from other threads.
    void queueInLoop(Functor cb);

    size_t queueSize() const;

    // timers

    /*
        EventLoop 新增了几个方便用户使用的定时器接口，这几个函数都转而调用
        TimerQueue::addTimer()。

        注意这几个 EventLoop 成员函数应该允许跨线程使用，比
        方说我想在某个 IO
       线程中执行超时回调。这就带来线程安全性方面的问题，muduo
        的解决办法不是加锁，而是把对 TimerQueue 的操作转移到 IO
       线程来进行，这会用到 EventLoop::runInLoop() 函数。
    */

    ///
    /// Runs callback at 'time'.
    /// Safe to call from other threads.
    ///
    TimerId runAt(Timestamp time, TimerCallback cb);
    ///
    /// Runs callback after @c delay seconds.
    /// Safe to call from other threads.
    ///
    TimerId runAfter(double delay, TimerCallback cb);
    ///
    /// Runs callback every @c interval seconds.
    /// Safe to call from other threads.
    ///
    TimerId runEvery(double interval, TimerCallback cb);
    ///
    /// Cancels the timer.
    /// Safe to call from other threads.
    ///
    void cancel(TimerId timerId);

    // internal usage
    void wakeup();
    void updateChannel(Channel* channel);
    void removeChannel(Channel* channel);
    bool hasChannel(Channel* channel);

    /*
        muduo 的接口设计会明确哪些成员函数是线程安全的，可以跨线程调用；哪
        些成员函数只能在某个特定线程调用（主要是 IO
       线程）。为了能在运行时检查这些 pre-condition，EventLoop 提供了
       isInLoopThread() 和 assertInLoopThread() 等函 数
    */
    // pid_t threadId() const { return threadId_; }
    void assertInLoopThread()
    {
        if (!isInLoopThread()) {
            abortNotInLoopThread();
        }
    }
    // 这里的 threadId_ 在实例对象被构造的时候就保存了构造该对象的线程的 ID
    // 这里判断是不是在同一个线程中。
    bool isInLoopThread() const { return threadId_ == CurrentThread::tid(); }
    // bool callingPendingFunctors() const { return callingPendingFunctors_; }
    bool eventHandling() const { return eventHandling_; }

    void setContext(const boost::any& context) { context_ = context; }

    const boost::any& getContext() const { return context_; }

    boost::any* getMutableContext() { return &context_; }

    /*
        既然每个线程至多有一个 EventLoop 对象，那么我们让 EventLoop 的 static
       成员 函数 getEventLoopOfCurrentThread() 返回这个对象。返回值可能为
       NULL，如果当前 线程不是 IO 线程的话。
    */
    static EventLoop* getEventLoopOfCurrentThread();

private:
    void abortNotInLoopThread();
    void handleRead();  // waked up
    void doPendingFunctors();

    void printActiveChannels() const;  // DEBUG

    typedef std::vector<Channel*> ChannelList;

    bool                        looping_; /* atomic */
    std::atomic<bool>           quit_;
    bool                        eventHandling_;          /* atomic */
    bool                        callingPendingFunctors_; /* atomic */
    int64_t                     iteration_;
    const pid_t                 threadId_;
    Timestamp                   pollReturnTime_;
    std::unique_ptr<Poller>     poller_;
    std::unique_ptr<TimerQueue> timerQueue_;
    int                         wakeupFd_;
    // unlike in TimerQueue, which is an internal class,
    // we don't expose Channel to client.
    std::unique_ptr<Channel> wakeupChannel_;
    boost::any               context_;

    // scratch variables
    ChannelList activeChannels_;
    Channel*    currentActiveChannel_;

    mutable MutexLock                     mutex_;
    std::vector<Functor> pendingFunctors_ GUARDED_BY(mutex_);
};

}  // namespace net
}  // namespace muduo

#endif  // MUDUO_NET_EVENTLOOP_H
