// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)

#include "muduo/net/EventLoopThread.h"

#include "muduo/net/EventLoop.h"

using namespace muduo;
using namespace muduo::net;

EventLoopThread::EventLoopThread(const ThreadInitCallback& cb,
                                 const string&             name)
    : loop_(NULL),
      exiting_(false),
      thread_(std::bind(&EventLoopThread::threadFunc, this), name),
      mutex_(),
      cond_(mutex_),
      callback_(cb)
{
}

EventLoopThread::~EventLoopThread()
{
    exiting_ = true;
    if (loop_ !=
        NULL)  // not 100% race-free, eg. threadFunc could be running callback_.
    {
        // still a tiny chance to call destructed object, if threadFunc exits
        // just now. but when EventLoopThread destructs, usually programming is
        // exiting anyway.
        loop_->quit();
        thread_.join();
    }
}

/*
    EventLoopThread 会启动自己的线程，并在其中运行 EventLoop::loop()。其中
    关键的 startLoop() 函数定义如下，这个函数会返回新线程中 EventLoop 对象的地
    址，因此用条件变量来等待线程的创建与运行。
*/
EventLoop* EventLoopThread::startLoop()
{
    assert(!thread_.started());
    thread_.start();

    EventLoop* loop = NULL;
    {
        MutexLockGuard lock(mutex_);
        while (loop_ == NULL) {
            cond_.wait();
        }
        loop = loop_;
    }

    return loop;
}

void EventLoopThread::threadFunc()
{
    EventLoop loop;

    if (callback_) {
        callback_(&loop);
    }

    {
        MutexLockGuard lock(mutex_);
        loop_ = &loop;
        cond_.notify();
    }

    loop.loop();

/*
    由于 EventLoop 的生命期与线程主函数的作用域相同，因此在 threadFunc() 退
    出之后这个指针就失效了。好在服务程序一般不要求能安全地退出（§9.2.2），这应该
    不是什么大问题。
*/
    // assert(exiting_);
    MutexLockGuard lock(mutex_);
    loop_ = NULL;
}
