// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)
//
// This is a public header file, it must only include public header files.

#ifndef MUDUO_NET_TCPCONNECTION_H
#define MUDUO_NET_TCPCONNECTION_H

#include <boost/any.hpp>
#include <memory>

#include "muduo/base/StringPiece.h"
#include "muduo/base/Types.h"
#include "muduo/base/noncopyable.h"
#include "muduo/net/Buffer.h"
#include "muduo/net/Callbacks.h"
#include "muduo/net/InetAddress.h"

// struct tcp_info is in <netinet/tcp.h>
struct tcp_info;

namespace muduo {
namespace net {

class Channel;
class EventLoop;
class Socket;

/*
    TcpConnection class 可谓是 muduo 最核心也是最复杂的 class，它的头文件和源
    文件一共有 450 多行，是 muduo 最大的 class。

    TcpConnection 是 muduo 里唯一默认使用 shared_ptr 来管理的 class，也是唯
    一继承 enable_shared_from_this 的 class，这源于其模糊的生命期。

    TcpConnection 没有可供用户使用的函数。

    TcpConnection 使用 Channel 来获得 socket 上的 IO 事件，它
    会自己处理 writable 事件，而把 readable 事件通过 MessageCallback
    传达给客户。 TcpConnection 拥有 TCP socket，它的析构函数会 close(fd)（在
    Socket 的析构函数 中发生）。

    注意 TcpConnection 表示的是“一次 TCP 连接”，它是不可再生的，一旦连接断
    开，这个 TcpConnection 对象就没啥用了。另外 TcpConnection 没有发起连接的功能，
    其构造函数的参数是已经建立好连接的 socket fd （无论是 TcpServer 被动接受还是
    TcpClient 主动发起），因此其初始状态是 kConnecting。
*/

///
/// TCP connection, for both client and server usage.
///
/// This is an interface class, so don't expose too much details.
class TcpConnection : noncopyable,
                      public std::enable_shared_from_this<TcpConnection> {
public:
    /// Constructs a TcpConnection with a connected sockfd
    ///
    /// User should not create this object.
    TcpConnection(EventLoop* loop, const string& name, int sockfd,
                  const InetAddress& localAddr, const InetAddress& peerAddr);
    ~TcpConnection();

    EventLoop*         getLoop() const { return loop_; }
    const string&      name() const { return name_; }
    const InetAddress& localAddress() const { return localAddr_; }
    const InetAddress& peerAddress() const { return peerAddr_; }
    bool               connected() const { return state_ == kConnected; }
    bool               disconnected() const { return state_ == kDisconnected; }
    // return true if success.
    bool   getTcpInfo(struct tcp_info*) const;
    string getTcpInfoString() const;

    // void send(string&& message); // C++11
    void send(const void* message, int len);
    void send(const StringPiece& message);
    // void send(Buffer&& message); // C++11
    void send(Buffer* message);  // this one will swap data
    void shutdown();             // NOT thread safe, no simultaneous calling
    // void shutdownAndForceCloseAfter(double seconds); // NOT thread safe, no
    // simultaneous calling
    void forceClose();
    void forceCloseWithDelay(double seconds);
    void setTcpNoDelay(bool on);
    // reading or not
    void startRead();
    void stopRead();
    bool isReading() const
    {
        return reading_;
    };  // NOT thread safe, may race with start/stopReadInLoop

    void setContext(const boost::any& context) { context_ = context; }

    const boost::any& getContext() const { return context_; }

    boost::any* getMutableContext() { return &context_; }

    void setConnectionCallback(const ConnectionCallback& cb)
    {
        connectionCallback_ = cb;
    }

    void setMessageCallback(const MessageCallback& cb)
    {
        messageCallback_ = cb;
    }

    void setWriteCompleteCallback(const WriteCompleteCallback& cb)
    {
        writeCompleteCallback_ = cb;
    }

    void setHighWaterMarkCallback(const HighWaterMarkCallback& cb,
                                  size_t                       highWaterMark)
    {
        highWaterMarkCallback_ = cb;
        highWaterMark_ = highWaterMark;
    }

    /// Advanced interface
    Buffer* inputBuffer() { return &inputBuffer_; }

    Buffer* outputBuffer() { return &outputBuffer_; }

/*
    TcpConnection class 也新增了 CloseCallback 事件回调，但是这个回调是给 TcpServer 和 TcpClient 用的，
    用于通知它们移除所持有的 TcpConnectionPtr，这不是给普通用户用的，普通用户继续使用 ConnectionCallback。
*/
    /// Internal use only.
    void setCloseCallback(const CloseCallback& cb) { closeCallback_ = cb; }

    // called when TcpServer accepts a new connection
    void connectEstablished();  // should be called only once
    // called when TcpServer has removed me from its map
    void connectDestroyed();  // should be called only once

private:
    enum StateE { kDisconnected, kConnecting, kConnected, kDisconnecting };
    void handleRead(Timestamp receiveTime);
    void handleWrite();
    void handleClose();
    void handleError();
    // void sendInLoop(string&& message);
    void sendInLoop(const StringPiece& message);
    void sendInLoop(const void* message, size_t len);
    void shutdownInLoop();
    // void shutdownAndForceCloseInLoop(double seconds);
    void        forceCloseInLoop();
    void        setState(StateE s) { state_ = s; }
    const char* stateToString() const;
    void        startReadInLoop();
    void        stopReadInLoop();

    EventLoop*   loop_;
    const string name_;
    StateE       state_;  // FIXME: use atomic variable
    bool         reading_;
    // we don't expose those classes to client.
    std::unique_ptr<Socket>  socket_;
    std::unique_ptr<Channel> channel_;
    const InetAddress        localAddr_;
    const InetAddress        peerAddr_;
    ConnectionCallback       connectionCallback_;
    MessageCallback          messageCallback_;
    WriteCompleteCallback    writeCompleteCallback_;
    HighWaterMarkCallback    highWaterMarkCallback_;
    CloseCallback            closeCallback_;
    size_t                   highWaterMark_;
    Buffer                   inputBuffer_;
    Buffer     outputBuffer_;  // FIXME: use list<Buffer> as output buffer.
    boost::any context_;
    // FIXME: creationTime_, lastReceiveTime_
    //        bytesReceived_, bytesSent_
};

typedef std::shared_ptr<TcpConnection> TcpConnectionPtr;

}  // namespace net
}  // namespace muduo

#endif  // MUDUO_NET_TCPCONNECTION_H
