// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)
//
// This is an internal header file, you should not include this.

#ifndef MUDUO_NET_TIMERQUEUE_H
#define MUDUO_NET_TIMERQUEUE_H

#include <set>
#include <vector>

#include "muduo/base/Timestamp.h"
#include "muduo/net/Callbacks.h"
#include "muduo/net/Channel.h"

namespace muduo {
namespace net {

class EventLoop;
class Timer;
class TimerId;

/*
    值得一提的是 TimerQueue 的数据结构的选择，TimerQueue 需要高效地组织目前
    尚未到期的 Timer，能快速地根据当前时间找到已经到期的 Timer，也要能高效地添
    加和删除 Timer。最简单的 TimerQueue 以按到期时间排好序的线性表为数据结构，
    muduo 最早也是用这种结构。这种结构的常用操作都是线性查找，复杂度是 O(N)。

    另一种常用做法是二叉堆组织优先队列（libev 用的是更高效的 4-heap），这种
    做法的复杂度降为 O(log N)，但是 C++ 标准库的 make_heap() 等函数不能高效地删
    除 heap 中间的某个元素，需要我们自己实现（令 Timer 记住自己在 heap 中的位置）。
    还有一种做法是使用二叉搜索树（例如 std::set/std::map），把 Timer 按到期
    时间先后排好序。操作的复杂度仍然是 O(log N)，不过 memory locality 比 heap 要
    差一些，实际速度可能略慢。但是我们不能直接用 map<Timestamp, Timer＊>，因为这
    样无法处理两个 Timer 到期时间相同的情况。有两个解决方案，一是用 multimap 或
    multiset，二是设法区分 key。muduo 现在采用的是第二种做法，这样可以避免使
    用不常见的 multimap class。具体来说，以 pair<Timestamp, Timer＊> 为 key，这样即
    便两个 Timer 的到期时间相同，它们的地址也必定不同。

*/

///
/// A best efforts timer queue.
/// No guarantee that the callback will be on time.
///

/*
    TimerQueue 使用了一个 Channel 来观察 timerfd_ 上的 readable 事件。
    TimerQueue 是 EventLoop 的定时器功能。
    TimerQueue 的成员函数只能在其所属的 IO 线程调用，因此不必加锁。

    传统的Reactor 通过控制 select(2) 和 poll(2) 的等待时间来实现定时，而现在在 Linux 中
    有了 timerfd，我们可以用和处理 IO 事件相同的方式来处理定时，代码的一致性更好。

    muduo 的定时器功能由三个 class 实现，TimerId、Timer、TimerQueue，用户只
    能看到第一个 class，另外两个都是内部实现细节。
*/
class TimerQueue : noncopyable {
public:
    explicit TimerQueue(EventLoop* loop);
    ~TimerQueue();

/*
    addTimer() 是供 EventLoop 使用的，
    EventLoop 会把它封装为更好用的 runAt()、runAfter()、runEvery() 等函数。
*/
    ///
    /// Schedules the callback to be run at given time,
    /// repeats if @c interval > 0.0.
    ///
    /// Must be thread safe. Usually be called from other threads.
    TimerId addTimer(TimerCallback cb, Timestamp when, double interval);

    void cancel(TimerId timerId);

private:
    // FIXME: use unique_ptr<Timer> instead of raw pointers.
    // This requires heterogeneous comparison lookup (N3465) from C++14
    // so that we can find an T* in a set<unique_ptr<T>>.
    typedef std::pair<Timestamp, Timer*> Entry;
    typedef std::set<Entry>              TimerList;
    typedef std::pair<Timer*, int64_t>   ActiveTimer;
    typedef std::set<ActiveTimer>        ActiveTimerSet;

    void addTimerInLoop(Timer* timer);
    void cancelInLoop(TimerId timerId);
    // called when timerfd alarms
    void handleRead();
    // move out all expired timers
    std::vector<Entry> getExpired(Timestamp now);
    void               reset(const std::vector<Entry>& expired, Timestamp now);

    bool insert(Timer* timer);

    EventLoop* loop_;
    const int  timerfd_;
    Channel    timerfdChannel_;
    // Timer list sorted by expiration
    TimerList timers_;

    // for cancel()
    ActiveTimerSet activeTimers_;
    bool           callingExpiredTimers_; /* atomic */
    ActiveTimerSet cancelingTimers_;
};

}  // namespace net
}  // namespace muduo
#endif  // MUDUO_NET_TIMERQUEUE_H
