// Copyright 2010, Shuo Chen.  All rights reserved.
// http://code.google.com/p/muduo/
//
// Use of this source code is governed by a BSD-style license
// that can be found in the License file.

// Author: Shuo Chen (chenshuo at chenshuo dot com)

#include "muduo/net/poller/PollPoller.h"

#include <assert.h>
#include <errno.h>
#include <poll.h>

#include "muduo/base/Logging.h"
#include "muduo/base/Types.h"
#include "muduo/net/Channel.h"

using namespace muduo;
using namespace muduo::net;

PollPoller::PollPoller(EventLoop* loop) : Poller(loop) {}

PollPoller::~PollPoller() = default;

Timestamp PollPoller::poll(int timeoutMs, ChannelList* activeChannels)
{
    /*
        Poller::poll() 是 Poller 的核心功能，它调用 poll(2) 获得当前活动的 IO 事
        件，然后填充调用方传入的 activeChannels，并返回 poll(2) return 的时刻。

        ::poll(...): 调用全局命名空间中的 poll 函数，对给定的文件描述符进行轮询。
        将 poll 函数的返回值存储在 numEvents 中，该返回值表示发生事件的文件描述符数量。
        timeoutMs 是轮询的超时时间。
    */
    // XXX pollfds_ shouldn't change
    int numEvents = ::poll(&*pollfds_.begin(), pollfds_.size(), timeoutMs);
    int savedErrno = errno;
    Timestamp now(Timestamp::now());
    if (numEvents > 0) {
        LOG_TRACE << numEvents << " events happened";
        fillActiveChannels(numEvents, activeChannels);
    } else if (numEvents == 0) {
        LOG_TRACE << " nothing happened";
    } else {
        /*
            savedErrno != EINTR: 在 poll 函数被信号中断（EINTR）的情况下，poll 函数返回 -1，
            但并不是真正的错误，而是表示等待被中断。所以，这个条件判断是否发生了信号中断。
            如果 poll 函数返回了 EINTR，则将 errno 设置为之前保存的错误码 savedErrno，以便进一步处理。
        */
        if (savedErrno != EINTR) {
            errno = savedErrno;
            LOG_SYSERR << "PollPoller::poll()";
        }
    }
    return now;
}

/*
    fillActiveChannels() 遍历 pollfds_，找出有活动事件的 fd，把它对应的 Channel
    填入 activeChannels。 这个函数的复杂度是 O(N)，其中 N 是 pollfds_ 的长度，
    即文件描述符数目。为了提前结束循环，每找到一个活动 fd 就递减 numEvents，这样
    当 numEvents 减为 0 时表示活动 fd 都找完了，不必做无用功。当前活动事件
    revents 会保存在 Channel 中，供 Channel::handleEvent() 使用。
    注意这里我们不能一边遍历 pollfds_，一边调用 Channel::handleEvent()，因
    为后者会添加或删除 Channel，从而造成 pollfds_ 在遍历期间改变大小，这是非常危
    险的。另外一个原因是简化 Poller 的职责，它只负责 IO multiplexing，不负责事件
    分发（dispatching）。这样将来可以方便地替换为其他更高效的 IO multiplexing 机
    制，如 epoll(4)。
*/
void PollPoller::fillActiveChannels(int          numEvents,
                                    ChannelList* activeChannels) const
{
    for (PollFdList::const_iterator pfd = pollfds_.begin();
         pfd != pollfds_.end() && numEvents > 0; ++pfd) {
        if (pfd->revents > 0) {
            --numEvents;
            ChannelMap::const_iterator ch = channels_.find(pfd->fd);
            assert(ch != channels_.end());
            Channel* channel = ch->second;
            assert(channel->fd() == pfd->fd);
            channel->set_revents(pfd->revents);
            // pfd->revents = 0;
            activeChannels->push_back(channel);
        }
    }
}

/*
    Poller::updateChannel() 的主要功能是负责维护和更新 pollfds_ 数组。添加新
    Channel 的复杂度是 O(log N)，更新已有的 Channel 的复杂度是 O(1)，因为
    Channel 记住了自己在 pollfds_ 数组中的下标，因此可以快速定位。removeChannel()
    的复杂 度也将会是 O(log N)。这里用了大量的 assert 来检查 invariant
*/
void PollPoller::updateChannel(Channel* channel)
{
    Poller::assertInLoopThread();
    LOG_TRACE << "fd = " << channel->fd() << " events = " << channel->events();
    if (channel->index() < 0) {
        // a new one, add to pollfds_
        assert(channels_.find(channel->fd()) == channels_.end());
        struct pollfd pfd;
        pfd.fd = channel->fd();
        pfd.events = static_cast<short>(channel->events());
        pfd.revents = 0;
        pollfds_.push_back(pfd);
        int idx = static_cast<int>(pollfds_.size()) - 1;
        channel->set_index(idx);
        channels_[pfd.fd] = channel;
    } else {
        // update existing one
        assert(channels_.find(channel->fd()) != channels_.end());
        assert(channels_[channel->fd()] == channel);
        int idx = channel->index();
        assert(0 <= idx && idx < static_cast<int>(pollfds_.size()));
        struct pollfd& pfd = pollfds_[idx];
        assert(pfd.fd == channel->fd() || pfd.fd == -channel->fd() - 1);
        pfd.fd = channel->fd();
        pfd.events = static_cast<short>(channel->events());
        pfd.revents = 0;
        if (channel->isNoneEvent()) {
            /*
                在这个特定的情况下，为了标记一个不感兴趣的 Channel，通过对文件描述符进行特殊处理，将其设置为 -channel->fd() - 1，
                以便在后续的轮询操作中能够识别并跳过这个文件描述符。
                
                -1 这并不是对文件描述符的通常操作，而是一种特定的技巧用于实现某种标记或标志的效果。将文件描述符设置为负数并减去1，
                通常是为了避免与合法的文件描述符产生冲突。在Linux系统中，通常情况下合法的文件描述符是非负整数。
            */
            // ignore this pollfd
            pfd.fd = -channel->fd() - 1;
        }
    }
}

void PollPoller::removeChannel(Channel* channel)
{
    Poller::assertInLoopThread();
    LOG_TRACE << "fd = " << channel->fd();
    assert(channels_.find(channel->fd()) != channels_.end());
    assert(channels_[channel->fd()] == channel);
    assert(channel->isNoneEvent());
    int idx = channel->index();
    assert(0 <= idx && idx < static_cast<int>(pollfds_.size()));
    const struct pollfd& pfd = pollfds_[idx];
    (void)pfd;
    assert(pfd.fd == -channel->fd() - 1 && pfd.events == channel->events());
    size_t n = channels_.erase(channel->fd());
    assert(n == 1);
    (void)n;
    if (implicit_cast<size_t>(idx) == pollfds_.size() - 1) {
        pollfds_.pop_back();
    } else {
        int channelAtEnd = pollfds_.back().fd;
        iter_swap(pollfds_.begin() + idx, pollfds_.end() - 1);
        if (channelAtEnd < 0) {
            channelAtEnd = -channelAtEnd - 1;
        }
        channels_[channelAtEnd]->set_index(idx);
        pollfds_.pop_back();
    }
}
